/* audioproc.cpp - part of MUSE
 * audio processing core routines
 */

#include "generic.h"

static short work1[32768], work2[32768];

void make_mono(short *stereo, short *mono, unsigned int num)
{
    unsigned int i;
    for(i=0;i<num;i++)
        mono[i]=(short)(((int)stereo[i<<1]+(int)stereo[(i<<1)+1])>>1);
} /* make_mono */

void make_stereo(short *mono, short *stereo, unsigned int num)
{
    unsigned int i;
    for(i=0;i<num;i++)
        stereo[1<<1]=stereo[(i<<1)+1]=mono[i];
}/* make_stereo */

void sep_stereo(short *stereo, short *left, short *right, unsigned int num)
{
    unsigned int i;
    for(i=0;i<num;i++)
    {
       	left[i]=stereo[i<<1];
       	right[i]=stereo[(i<<1)+1];
    }
} /* sep_stereo  */

void mix_stereo(short *left, short *right, short *stereo, unsigned int num)
{
    unsigned int i;
    for(i=0;i<num;i++)
    {
        stereo[i<<1]=left[i];
	stereo[(i<<1)+1]=right[i];
    }
} /* mix_stereo */

void int_compress_samples(short *in, short *out,unsigned int num,unsigned int grain)
{
    unsigned int i,j;
    int smpl;
	
    for(i=0;i<num;i++)
    {
       	smpl=in[i*grain];
	for(j=1;j<grain;j++) smpl+=in[i*grain+j];
	out[i]=(short)(smpl/grain);

    }
} /* int_compress_samples */

void int_stretch_samples(short *in,short *out,unsigned int num,unsigned int grain)
{
    unsigned int i,j;

    for(i=0;i<num;i++)
       	for(j=0;j<grain;j++)
            out[i*grain+j]=in[i];
	
} /* int_stretch_samples  */

void compress_samples(short *in, short *out,unsigned int num1,unsigned int num2)
{
    float step=(float)(num1)/num2;
    unsigned int i;

    for(i=0;i<num2;i++) out[i]=in[(int)(i*step)];
} /* compress_samples */

void stretch_samples(short *in, short *out,unsigned int num1,unsigned int num2)
{
    float step=(float)(num1)/num2;
    unsigned int i;

    for(i=0;i<num2;i++) out[i]=in[(int)(i*step)];
}/* stretch_samples */

/* changes the length of a buffer of samples uses fast shortcuts where possible */
void change_time(short *in, short *out, unsigned int num_in, unsigned int num_out){
    unsigned int i;
	
    if(num_in>num_out)
    {
        /* compress samples */
        if(num_in==(num_out<<1))
        {
            int_compress_samples(in,out,num_out,2);
        }
        else if( num_in==(num_out<<2))
        {
            int_compress_samples(in,out,num_out,4);
        }
        else
        {
            compress_samples(in,out,num_in,num_out);
        }
    }
    else if(num_in<num_out)
    {
        /* stretch samples */
        if((num_in<<1)==num_out)
        {
            int_stretch_samples(in,out,num_in,2);
        }
        else if ((num_in<<2)==num_out)
        {
            int_stretch_samples(in,out,num_in,4);
        }
        else
        {
            stretch_samples(in,out,num_in,num_out);
        }
    }
    else
    {
        for(i=0;i<num_out;i++)
            out[i]=in[i];
    }
} /* change_time   */

void convert_audio(short *in,short *out,int ch1,int ch2,int n1,int n2)
{
    int i;

    if((ch1==ch2) &&(n1==n2))
    {
      int gh = ch1*n1;
      for(i=0;i<gh;i++)
	out[i]=in[i]; 
        return;
    }

    if(ch2==1){
        /* output is mono */
        if(ch1==2){
            /* input is stereo */
            make_mono(in,work1,n1);
            change_time(work1,out,n1,n2);
        }
        else
        {
            /* input is mono */
            change_time(in,out,n1,n2);
        }
    }
    else
    {
        /* output is stereo */
        if(ch1==2){
            /* input is stereo */
            sep_stereo(  in,    work1, work2, n1    );
            change_time( work1, in,    n1,    n2    );
            change_time( work2, work1, n1,    n2    );
            mix_stereo(  in,    work1, out,   n2    );
        }
        else
        {
            /* input is mono */
            change_time(in,work1,n1,n2);
            make_stereo(work1,out,  n2);
        }
    }
} /* convert_audio */

void clip_audio(int *pcm, short *clipped, int chunks) {
  int c;

  for(c=0;c<chunks;c++) {
    if((pcm[c]<CLIP)&&(pcm[c]>-CLIP))
      clipped[c]=(short)(pcm[c]);
    else if (pcm[c]>0)
      clipped[c]=MAX_SAMPLE-(CLIP_A/(CLIP_B+pcm[c]));
    else
      clipped[c]=-(MAX_SAMPLE-(CLIP_A/(CLIP_B-pcm[c])));
  }
}

int get_level(short *pcm, int chunks) {
  int c;
  int max = 0;

  for(c=0;c<chunks;c++)
    if(pcm[c]>max)
      max = pcm[c];

  return(max);
}

short *invert(short *pcm, int chunks) {
  short t;
  short *a, *b;

  a = pcm;
  b = pcm+chunks;
  
  while(b>0) {
    t = *a;
    *a = *b;
    *b = t;
    a++;
    b--;
  }

  return(b);
}
