/* inchannels.cpp - MuSE
   (c)2000 Denis Roio aka jaromil
   see COPYING for licensing stuff

   different classes for different IN channels
   they are instantiated and used by the Stream_mixer class (jmixer.cpp)
*/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include "audioproc.h"
#include "inchannels.h"

/* xaudio includes */
#include "xaudio/file_input.h"
#include "xaudio/mpeg_codec.h"

extern "C" {
#include "mpeg.h"
}

static short inbuf[BUF_SIZE*40];

/* ----- MP3 CHANNEL ----- */

Mp3Channel::Mp3Channel() {
  volume = 1.0;
  speed = 1.0;
  rate = 0;
  channels = 0;
  on = false;
  end = false;
  stream[0] = 0;
  stream[1] = 0;
  pid = 0;
}

Mp3Channel::~Mp3Channel() {
  if(on)
    stop();
}

void Mp3Channel::start(int plpos) {
  if(pid) stop();
  name = &playlist[plpos][0];
  spawn();
  on = true;
  end = false;
}

void Mp3Channel::stop() {
  if(pid>0) {

    if(!end)
      kill(pid,3);

    waitpid(pid,NULL,0x0);
    
    close(stream[0]);
    close(stream[0]);
    
    on = false;
    end = false;
    pid = 0;
    stream[0] = 0;
    stream[1] = 1;
  } else {
    on = false;
    end = false;
  }

  memset(gotin,0,BUF_SIZE);
}

int Mp3Channel::spawn() {
  int strm[2], trate, tchannels;
  pid_t mypid;

  trate=format_of(name,&tchannels);
  rate=trate;
  channels=tchannels;

  pipe(strm);
  mypid=fork();
  if(mypid) {
    pid = mypid;
    stream[0] = strm[0];
    stream[1] = strm[1];
    return(0);
  } else {
    close(strm[0]);
    close(1);
    dup(strm[1]);
    close(strm[1]);

    /* QUAAAAAA */
    close(*dsp);
    execlp("mpg123","mpg123","-sq"/*,"-n 300"*/,name,NULL);
  }
  return -1;
}

void Mp3Channel::ends() {

  kill(pid,3); /* just to be sure */

  waitpid(pid,NULL,0x0);
  
  close(stream[0]);
  close(stream[1]);
  
  on = false;
  end = false;
  pid = 0;
  stream[0] = 0;
  stream[1] = 0;
}

unsigned int Mp3Channel::get_audio() {
  int i;
  unsigned int num_samples = 0;
  int opt;

  //  if(on) {

  num_samples=(int)((rate*BUF_SIZE*speed)/(sample_rate<<1)); /* for mono would be /(sample_rate) */
  
  opt = (num_samples*channels)<<1;
  
  i=read(stream[0],inbuf,opt);
  
  /* if we reach the end of the stream then ...
     Ok. It's a bug. it will be listed into the known bugs
     and it will be hopefully fixed. */
  if(i<opt) {
    /* for padding: for(;i<opt;i++) inbuf[i]=0; */
    end=true;
  }
  
  /* now convert the audio into something useful */
  convert_audio(inbuf,gotin,channels,2,num_samples,BUF_SIZE>>1);
  /* for mono would be:
     convert_audio(inbuf,stream_channel[ch].inbuf,stream_channel[ch].channels,1,num_samples,BUF_SIZE); */
  
  return num_samples;
} /* get_audio */

void Mp3Channel::get_playlist(char *plfile) {
  FILE *plist;
  int c = 0;
  char *p;

  plist = fopen(plfile,"r");
  if(plist == NULL) {
    fprintf(stderr,"Playlist file not found\n");
    fprintf(stderr,"to create a playlist file just do:\n");
    fprintf(stderr,"find /your/mp3/dir -name '*.mp3' -print >playlist\n");
    exit(0);
  }

  while (fgets(&playlist[c][0],MAX_PATH_SIZE,plist) != NULL) {
    p = &playlist[c][0];
    while(*p!='\n') p++;
    *p = '\0';
    c++;
  }
  fclose(plist);
  
  playlist_len = c;

}

/* ----- MP3 CHANNEL ----- */

/* ----- XAudio MP3 CHANNEL ----- */

XaudioChannel::XaudioChannel() {

  decoder_new(&decoder);

   /* register mpeg audio codec */
    {
        XA_CodecModule module;

        mpeg_codec_module_register(&module);
        decoder_codec_module_register(decoder, &module);
    }

    /* register the file input module */
    {
        XA_InputModule module;

        file_input_module_register(&module);
        decoder_input_module_register(decoder, &module);
    }

    input = decoder->input;
    status = decoder->status;
    buffer = decoder->output_buffer;

    volume = 1.0;
    timeline = 0.0;
    opened = false;
    
}

XaudioChannel::~XaudioChannel() {
  if(opened)
    decoder_input_delete(decoder);
}

bool XaudioChannel::set(int plpos) {
  int res;

  /* if an input was already created and opened
     then is automatically closed and deleted by decoder_input_new() */

  res = decoder_input_new(decoder,&playlist[plpos][0],XA_DECODER_INPUT_AUTOSELECT);
  opened = false;

  if(res != XA_SUCCESS)
    return(false);
    
  selected = plpos;
  return(true);
}

float XaudioChannel::get_audio() {
  int res;
  res = decoder_decode(decoder,NULL);
  if(res == XA_SUCCESS)
    return(status->position);

  if(res == XA_ERROR_INPUT_EOF) {/* the input stream ended */
    stop(); /* should do next thing: loop, nextsong or stop */
    return(2.0);
  }
  
  fprintf(stderr,"XA: i have a bad feeling\n");
  return(3.0);
}

bool XaudioChannel::play() {
  int res;
  
  res = decoder_input_open(decoder);
  if(res != XA_SUCCESS)
    return(false);

  fprintf(stderr,"MPEG layer %u\n",decoder->status->info.layer);
  fprintf(stderr,"bitrate %u\n",decoder->status->info.bitrate);
  fprintf(stderr,"freq %u\n",decoder->status->info.frequency);
  
  on = true;
  opened = true;
  return(true);
}

bool XaudioChannel::stop() {
  int res;
  
  res = decoder_input_close(decoder);
  if (res != XA_SUCCESS)
    return(false);

  on = false;
  opened = false;
  return(true);
}

bool XaudioChannel::pos(float position) {
  int res;

  res = decoder_input_seek_to_position(decoder,position);
  if (res!=XA_SUCCESS)
    return(false);
  
  return(true);
}

bool XaudioChannel::get_playlist(char *pl) {
  FILE *plist;
  int c = 0;
  char *p;

  plist = fopen(pl,"r");
  if(plist == NULL)
    return(false);

  while (fgets(&playlist[c][0],MAX_PATH_SIZE,plist) != NULL) {
    p = &playlist[c][0];
    while(*p!='\n') p++;
    *p = '\0';
    c++;
  }

  fclose(plist);
  
  playlist_len = c;
  return(true);
}

/* ----- LiveIN DSP Channel ----- */

LiveIn::LiveIn(int smpr, int chans, int *thedsp) {
  dsp = thedsp;
  sample_rate = smpr;
  channels = chans;

  num_samples=(int)((sample_rate*BUF_SIZE)/(SAMPLE_RATE<<1));
  opt = (num_samples*channels)<<1;
}

LiveIn::~LiveIn() { };

bool LiveIn::get_audio() {
  read(*dsp,inbuf,opt);

  convert_audio(inbuf,gotin,channels,2,num_samples,BUF_SIZE>>1);

  return(true);
}
